import npyscreen

from py_slack_term.lib.UI.forms import SlackConversationsWindowForm
from py_slack_term.lib.UI.themes import UITheme


class SlackApplication(npyscreen.NPSAppManaged):
    def __init__(self, *args, slack_client: object=None, config: object=None, **kwargs):
        npyscreen.NPSAppManaged.STARTING_FORM = 'conversations_window'

        self.slack_client = slack_client
        self.config = config

        super(SlackApplication, self).__init__(*args, **kwargs)

        self.conversations_window = None

    def onStart(self) -> None:
        npyscreen.setTheme(UITheme)
        self.conversations_window = self.addForm('conversations_window',
                                                 SlackConversationsWindowForm,
                                                 name='Slack Terminal',
                                                 slack_client=self.slack_client,
                                                 config=self.config)

    def stop(self) -> None:
        self.conversations_window.stop()


import npyscreen
from npyscreen.wgmultiline import MORE_LABEL


class PatchedBufferPager(npyscreen.BufferPager):
    def clear_buffer(self, *args, **kwargs):
        """
        compatibility with non pythonic code in library
        """
        self.clearBuffer(*args, **kwargs)

    # TODO: this is a monkey patch of the base class Pager
    # TODO: this method can be removed when https://github.com/npcole/npyscreen/pull/60 is merged
    def update(self, clear=True):
        #we look this up a lot. Let's have it here.
        if self.autowrap:
            # this is the patch     V----------------------------V
            self.setValuesWrap(list(self.display_value(l) for l in self.values))

        if self.center:
            self.centerValues()

        display_length = len(self._my_widgets)
        values_len = len(self.values)

        if self.start_display_at > values_len - display_length:
            self.start_display_at = values_len - display_length
        if self.start_display_at < 0:
            self.start_display_at = 0

        indexer = 0 + self.start_display_at
        for line in self._my_widgets[:-1]:
            self._print_line(line, indexer)
            indexer += 1

        # Now do the final line
        line = self._my_widgets[-1]

        if values_len <= indexer+1:
            self._print_line(line, indexer)
        else:
            line.value = MORE_LABEL
            line.highlight = False
            line.show_bold = False

        for w in self._my_widgets:
            # call update to avoid needless refreshes
            w.update(clear=True)
        # There is a bug somewhere that affects the first line.  This cures it.
        # Without this line, the first line inherits the color of the form when not editing. Not clear why.
        self._my_widgets[0].update()



import json
import threading
import time
import websocket

from py_slack_term.lib import Logger


class SlackRTMClient:
    def __init__(self, slack_client, callback):
        self.slack_client = slack_client
        self.callback: classmethod = callback
        self.logger = Logger(' ')
        self.url: str = None
        self.ws: websocket.WebSocketApp = None
        self.wst: threading.Thread = None

    def start(self) -> None:
        self.logger.log('starting RTM client')
        while not self.url:
            self.url = self.slack_client.rtm_connect()
            if not self.url:
                self.logger.log('error getting RealTimeMessaging URL. waiting 10 seconds...')
                time.sleep(10)

        self.ws = websocket.WebSocketApp(self.url, on_message=self.on_message, on_error=self.on_error)
        self.wst = threading.Thread(target=self.ws.run_forever)
        self.wst.daemon = True
        self.wst.start()
        self.logger.log('started RTM client')

    def on_message(self, _, message: str) -> None:
        data = json.loads(message)
        self.callback(data)

    def on_error(self, *args: list) -> None:
        self.logger.log(args)
        self.stop()
        self.start()

    def stop(self) -> None:
        self.logger.log('closing RTM client')
        if hasattr(self, 'ws'):
            self.ws.close(timeout=1)
            del self.ws
        if hasattr(self, 'wst'):
            del self.wst
        if hasattr(self, 'url'):
            self.url = None
        self.logger.log('closed RTM client')



import curses

import npyscreen

from ....lib.slack_client.API import Channel


class ChannelSelector(npyscreen.MultiLine):
    """
    TODO: Make this a tree heirarchy that allows you to have multiple organisations (npyscreen.MLTree)
    selected channel colour is set as "BOLD"
    """
    def __init__(self, *args, **kwargs):
        super(ChannelSelector, self).__init__(*args, **kwargs)

    def display_value(self, vl: Channel) -> str:
        prefix = '*' if vl.has_unread else ' '
        if vl.is_private:
            prefix += ' <>'
        elif vl.is_channel:
            prefix += '  #'
        else:
            prefix += '  @'
        return prefix + vl.name

    def h_select(self, ch) -> None:
        """
        returns the currently selected channel object
        :param ch:
        :return:
        """
        super(ChannelSelector, self).h_select(ch)
        self.parent.select_channel(self.values[self.value])

    def set_up_handlers(self):
        super(ChannelSelector, self).set_up_handlers()
        self.handlers.update({
            curses.KEY_RIGHT: self.h_exit_right,
            curses.KEY_LEFT: self.h_exit_left,
            ord('d'): self.leave_channel
        })

    def leave_channel(self, *args):
        cur_channel = self.values[self.cursor_line]
        cur_channel.leave()
        self.values.remove(cur_channel)
        self.display()


class BoxedChannelSelector(npyscreen.BoxTitle):
    _contained_widget = ChannelSelector

    def __init__(self, *args, **kwargs):
        self.name = 'Channels'
        super(BoxedChannelSelector, self).__init__(*args, **kwargs)

    def update_channels(self, in_channels) -> None:
        self.values = in_channels



import npyscreen

from py_slack_term.lib import Logger
from py_slack_term.lib.UI.widgets import BoxedChannelSelector, BoxedChannelMessages, BoxedMessageComposer
from py_slack_term.lib.UI.widgets.debug_widget import BoxedScreenLogger
from py_slack_term.lib.slack_client.API import Message, SlackApiClient
from py_slack_term.lib.slack_client.RTM import SlackRTMClient


class SlackConversationsWindowForm(npyscreen.FormBaseNew):
    def __init__(self, *args, slack_client=None, config=None, **kwargs):
        self.config = config
        self.logger = Logger('SlackWindowForm')
        self.slack_client = slack_client  # type: SlackApiClient
        super(SlackConversationsWindowForm, self).__init__(*args, **kwargs)
        self.channel_selector = None
        self.current_channel = None

        self.rtm_client = SlackRTMClient(self.slack_client, self.new_RTM_event)
        self.rtm_client.start()
        if self.config.debug:
            self.logger = self.screen_logger
        self.slack_client.logger = self.logger
        self.rtm_client.logger = self.logger
        self.channel_messages.typing_user_watchdog_thread.logger = self.logger

    def create(self):
        y, x = self.useable_space()

        if self.config.debug:
            self.channel_selector = self.add_widget(BoxedChannelSelector, max_width=(x // 4) -2)
            self.channel_messages = self.add_widget(BoxedChannelMessages,
                                                    relx=self.channel_selector.width + 2,
                                                    rely=self.channel_selector.rely,
                                                    max_height=y-8,
                                                    max_width=(x // 2)-2)  # type: BoxedChannelMessages
            self.message_composer = self.add_widget(BoxedMessageComposer, relx=self.channel_messages.relx, rely=y-6, max_height=4, max_width = self.channel_messages.width)
            self.screen_logger = self.add_widget(BoxedScreenLogger, relx=(x // 4 * 3), rely=self.channel_selector.rely)
        else:
            self.channel_selector = self.add_widget(BoxedChannelSelector, max_width=x // 5)
            self.channel_messages = self.add_widget(BoxedChannelMessages,
                                                    relx=self.channel_selector.width + 3,
                                                    rely=self.channel_selector.rely,
                                                    max_height=y-8)  # type: BoxedChannelMessages
            self.message_composer = self.add_widget(BoxedMessageComposer, relx=self.channel_messages.relx, rely=y-6, max_height=4)

        self.refresh_channels()

    def select_channel(self, ch):
        self.current_channel = ch
        self.channel_messages.set_channel(ch)
        self.channel_messages.clear_buffer()
        self.channel_messages.buffer(list(reversed(ch.fetch_messages())))
        self.current_channel.has_unread = False

    def refresh_channels(self):
        self.channel_selector.update_channels(self.slack_client.get_active_channels_im_in())

    def send_message(self):
        message = self.message_composer.value
        self.message_composer.clear_message()
        self.current_channel.post_message(msg=message)

    def new_RTM_event(self, event: dict):
        if event != {}:
            self.logger.log(event)

        event_type = event.get('type')
        if event_type == 'hello':
            if self.current_channel:
                self.select_channel(self.current_channel)
                self.channel_messages.display()

        if event_type == 'message':
            message = Message(self.slack_client, **event)
            if self.current_channel:
                if event.get('channel') == str(self.current_channel.id):
                    self.channel_messages.buffer([message])
                    self.current_channel.mark(message.ts)
                self.current_channel.has_unread = False
                if self.channel_selector:
                    self.channel_selector.display()

        elif event_type in ('channel_marked', 'im_marked', 'group_marked'):
            chan = self.slack_client.channels.get(event.get('channel'))
            if chan:
                chan.register_ts(event.get('ts'), as_read=True)
                if self.channel_selector:
                    self.channel_selector.display()

        elif event_type == 'user_typing':
            chan = self.slack_client.channels.get(event.get('channel'))
            user = event.get('user')
            chan.register_typing_user(user)
            self.channel_messages.typing_user_event()

    def stop(self):
        self.rtm_client.stop()


from .client import SlackApiClient
from .message import Message
from .user import User

from py_slack_term.lib.slack_client.API.channel import Channel

__all__ = [
    Channel,
    User,
    Message,
    SlackApiClient
]


import time

from .message import Message


class Channel:
    def __init__(self, client, **kwargs):
        self.client = client
        self.id = kwargs.get('id')
        self.name = kwargs.get('name') or self.client.users.get(kwargs.get('user')).name
        self.is_channel = kwargs.get('is_channel')
        self.created = kwargs.get('created')
        self.is_archived = kwargs.get('is_archived')
        self.is_general = kwargs.get('is_general')
        self.unlinked = kwargs.get('unlinked')
        self.creator = kwargs.get('creator')
        self.name_normalized = kwargs.get('name_normalized')
        self.is_shared = kwargs.get('is_shared')
        self.is_org_shared = kwargs.get('is_org_shared')
        self.is_member = kwargs.get('is_member')
        self.is_private = kwargs.get('is_private')
        self.is_mpim = kwargs.get('is_mpim')
        self.members = {u.get_name(): u for u in [self.client.users[m] for m in self.get_members()]}
        self.topic = kwargs.get('topic')
        self.purpose = kwargs.get('purpose')
        self.previous_names = kwargs.get('previous_names')
        self.num_members = kwargs.get('num_members')
        self.last_seen_ts = 0
        self.has_unread = True
        messages = self.fetch_messages(read=False)
        self.last_seen_ts = float(messages[0].ts)

        channel_info = self.get_info()
        last_read = channel_info.get('last_read')
        if last_read:
            self.register_ts(last_read, as_read=True)
        self.typing_users = {}

    def register_typing_user(self, user: str) -> None:
        self.typing_users[self.client.users[user]] = time.time()

    def register_ts(self, ts: float, *_, as_read: bool=False) -> None:
        if float(ts) >= float(self.last_seen_ts):
            if as_read:
                self.last_seen_ts = float(ts)
                self.has_unread = False
            elif float(ts) > float(self.last_seen_ts) and not as_read:
                self.has_unread = True
        elif not as_read:
            self.has_unread = False

    def get_info(self) -> dict:
        response = self.client.slackclient.api_call('conversations.info', channel=self.id)
        if response.get('ok'):
            return response.get('channel')

    def get_members(self) -> dict:
        response = self.client.slackclient.api_call('conversations.members', channel=self.id)
        if response.get('ok'):
            return response.get('members')

    def join(self) -> dict:
        response = self.client.slackclient.api_call('channels.join', channel=self.id)
        if response.get('ok'):
            return response

    def leave(self) -> dict:
        response = self.client.slackclient.api_call('conversations.leave', channel=self.id)
        if response.get('ok'):
            return response

    def post_message(self, msg: str, thread_ts: float=None, reply_broadcast: bool=False) -> dict:
        """
        https://api.slack.com/methods/chat.postMessage
        """
        # replace @annotated mentions with the correct escape sequence
        for notification in ('here', 'everyone', 'channel'):
            msg = msg.replace('@' + notification, '<!' + notification + '>')

        # replace @annotated user mentions with <@USERID> sequence
        for name, member in self.members.items():
            if '@' + name + ' ' in msg:
                msg = msg.replace('@' + name, '<@' + member.id + '>')

        return self.client.slackclient.api_call('chat.postMessage',
                                                channel=self.id,
                                                text=msg,
                                                link_names=True,
                                                as_user=True,
                                                thread_ts=thread_ts,
                                                reply_broadcast=reply_broadcast)

    def post_ephemeral_message(self, msg:str, user: str) -> dict:
        response = self.client.slackclient.api_call('chat.postEphemeral',
                                                channel=self.id,
                                                text=msg,
                                                user=user)
        if response.get('ok'):
            return response

    def delete_message(self, msg_ts: float) -> dict:
        response = self.client.slackclient.api_call('chat.delete',
                                                channel=self.id,
                                                ts=msg_ts)
        if response.get('ok'):
            return response

    def fetch_messages(self, read: bool=True) -> list:
        response = self.client.slackclient.api_call('conversations.history',
                                                    channel=self.id,
                                                    count=200)
        if response.get('ok'):
            messages = [Message(self.client, **message) for message in response.get('messages')]
            if len(messages) > 0:
                if read:
                    self.mark(messages[0].ts)
            return messages

    def mark(self, ts: float) -> None:
        if self.is_mpim:
            endpoint = 'mpim'
        elif self.is_private:
            endpoint = 'groups'
        elif self.is_channel:
            endpoint = 'channels'
        else:
            endpoint = 'im'
        self.client.slackclient.api_call(endpoint + '.mark',
                                         channel=self.id,
                                         ts=ts)



import curses

import npyscreen


class ScreenLogger(npyscreen.BufferPager):
    def __init__(self, *args, **kwargs):
        super(ScreenLogger, self).__init__(*args, **kwargs)
        self.autowrap = True

    def set_up_handlers(self):
        super(ScreenLogger, self).set_up_handlers()
        self.handlers.update({
            curses.KEY_LEFT: self.h_exit_left,
            curses.KEY_RIGHT: self.h_exit_right
        })


class BoxedScreenLogger(npyscreen.BoxTitle):
    _contained_widget = ScreenLogger

    def __init__(self, *args, **kwargs):
        self.name = "Debug"
        super(BoxedScreenLogger, self).__init__(*args, **kwargs)

    def log(self, msg):
        self.entry_widget.buffer([str(msg)])
        self.display()


import pprint

import time

from py_slack_term.lib import Config
from py_slack_term.lib.slack_client.API import SlackApiClient
from py_slack_term.lib.slack_client.RTM import SlackRTMClient


class DummyLogger:
    def log(self, msg):
        pprint.pprint(msg)


if __name__ == '__main__':
    """
    this is just a dummy testing stub. not executed when ran from cli
    """

    print = pprint.pprint
    config = Config()
    client = SlackApiClient(config)

    def api_test():
        pass
        #client.refresh_channel_list()
        #print(client.channels)
        #print([m.text for m in client.channels['admin'].fetch_messages()])

        #client.refresh_user_list()
        #print(client.users)
        for channel in client.channels.values():
            if channel.name == 'random':
                print(channel.members)

    def rtm_test():
        rtm_client = SlackRTMClient(client, print)
        rtm_client.logger = DummyLogger()
        rtm_client.start()
        while True:
            time.sleep(10)
            pass

    api_test()




import os
import yaml


class Config:
    config_path = os.path.expanduser('~') + '/.config/py_slack_term/'
    config_filename = 'config.yml'

    def __init__(self):
        if not os.path.exists(self.config_path):
            os.mkdir(self.config_path)
        if not os.path.isfile(self.config_path + self.config_filename):
            with open(self.config_path + self.config_filename, 'w+') as config_file:
                config = dict(
                    slacktoken=input("Please paste your slack API token: ")
                )
                config_file.write(yaml.dump(config))

        with open(self.config_path + self.config_filename) as config_file:
            config = yaml.load(config_file)
        self.token: str = config.get('slacktoken')
        self.debug: bool = True if config.get('debug') else False




import npyscreen


class UITheme(npyscreen.ThemeManager):
    default_colors = {
    'DEFAULT'     : 'WHITE_BLACK',
    'FORMDEFAULT' : 'WHITE_BLACK',
    'NO_EDIT'     : 'BLUE_BLACK',
    'STANDOUT'    : 'CYAN_BLACK',
    'CURSOR'      : 'WHITE_BLACK',
    'CURSOR_INVERSE': 'BLACK_WHITE',
    'LABEL'       : 'GREEN_BLACK',
    'LABELBOLD'   : 'BLACK_RED',
    'CONTROL'     : 'YELLOW_BLACK',
    'IMPORTANT'   : 'GREEN_BLACK',
    'SAFE'        : 'GREEN_BLACK',
    'WARNING'     : 'YELLOW_BLACK',
    'DANGER'      : 'RED_BLACK',
    'CRITICAL'    : 'BLACK_RED',
    'GOOD'        : 'GREEN_BLACK',
    'GOODHL'      : 'GREEN_BLACK',
    'VERYGOOD'    : 'BLACK_GREEN',
    'CAUTION'     : 'YELLOW_BLACK',
    'CAUTIONHL'   : 'BLACK_YELLOW',
    'BOLD'        : 'BLACK_GREEN'
    }


import curses

import npyscreen


class MessageComposer(npyscreen.MultiLineEdit):
    def __init__(self, *args, **kwargs):
        super(MessageComposer, self).__init__(*args, **kwargs)

    def set_up_handlers(self):
        super(MessageComposer, self).set_up_handlers()
        self.handlers.update({
            curses.KEY_ENTER: self.h_enter,
            curses.ascii.LF: self.h_enter
        })

    def h_enter(self, *args):
        self.parent.send_message()


class BoxedMessageComposer(npyscreen.BoxTitle):
    _contained_widget = MessageComposer

    def __init__(self, *args, **kwargs):
        self.name = 'Message'
        super(BoxedMessageComposer, self).__init__(*args, **kwargs)

    def clear_message(self):
        self.entry_widget.value = ''



from .channel_messages import BoxedChannelMessages
from .channel_selector import BoxedChannelSelector

from py_slack_term.lib.UI.widgets.message_composer import BoxedMessageComposer

__all__ = [
    BoxedMessageComposer,
    BoxedChannelMessages,
    BoxedChannelSelector
]


import pkg_resources
import argparse
import py_slack_term

from .lib import Config
from .lib.UI import SlackApplication
from py_slack_term.lib.slack_client.API import SlackApiClient

py_slack_term.User.ADMIN_USER_PREFIX = '[ADMIN]'
py_slack_term.User.PREFER_REAL_NAME = False

parser = argparse.ArgumentParser()
parser.add_argument('--version', action='store_true')
parser.add_argument('--debug', action='store_true')
args = parser.parse_args()


def main() -> None:
    debug = False
    if args.version:
        version = pkg_resources.get_distribution("py_slack_term").version
        print("\nVersion: py_slack_term: {}".format(version))
        return
    if args.debug:
        debug = True
    config = Config()
    if debug:
        config.debug = True
    slackclient = SlackApiClient(config)
    app = SlackApplication(slack_client=slackclient, config=config)
    try:
        app.run()
    except KeyboardInterrupt:
        app.stop()
        pass

if __name__ == "__main__":
    main()



from distutils.core import setup
import os

from setuptools import find_packages

VERSION = os.environ.get('TRAVIS_TAG') or '0.0.1-untagged'

setup(
    name='py_slack_term',
    version=VERSION,
    packages=find_packages(),
    url='https://github.com/chestm007/py_slack_terminal',
    license='GPL-2.0',
    author='max',
    author_email='chestm007@hotmail.com',
    description='Terminal based client for Slack',
    install_requires=[
        "slackclient",
        "npyscreen",
        "pyyaml",
        'websocket-client==0.48.0'
    ],
    entry_points="""
        [console_scripts]
        slack-term=py_slack_term.main:main
    """
)



from datetime import datetime


class Message:
    def __init__(self, client, **kwargs):
        self.client = client
        self.channel = self.client.channels.get(kwargs.get('channel'))
        self.user = client.users.get(kwargs.get('user'))
        self.text = kwargs.get('text') or ''
        self.type = kwargs.get('type')
        self.subtype = kwargs.get('subtype')
        self.ts = kwargs.get('ts')
        if self.channel:
            self.channel.register_ts(self.ts)

    def to_format_dict(self):
        return dict(
            user=' ' if self.subtype == 'bot_message' else self.user,
            text=self.text,
            type=self.type,
            subtype=self.subtype,
            ts=self.ts,
            time=datetime.fromtimestamp(float(self.ts))
        )


import curses
import re
import threading
import time
import npyscreen
from datetime import datetime

from py_slack_term.lib import Logger
from py_slack_term.lib.npyscreen_patch.buffer_pager import PatchedBufferPager
from ....lib.slack_client.API import Channel, Message


class ChannelMessages(PatchedBufferPager):
    mention_regex = re.compile("<@[A-Z0-9]+>")

    def __init__(self, *args, **kwargs):
        super(ChannelMessages, self).__init__(*args, **kwargs)
        self.autowrap = True

    def display_value(self, vl: Message) -> str:
        if isinstance(vl, Message):
            message_dict = vl.to_format_dict()
            text = str(message_dict.get('text'))
            if text is not None:
                match = re.search(self.mention_regex, text)
                if match:
                    # replace the <@XXXXXX> decorations, and leave only the user id
                    user_id = match.group().replace('<', '').replace('@', '').replace('>', '')
                    # replace userid with an '@' annotated username
                    message_dict['text'] = message_dict.get('text').replace(match.group(),
                                                                            '@' + vl.client.users.get(user_id).get_name())
                message_format = f"[{datetime.fromtimestamp(float(message_dict.get('ts'))):%d/%m-%H:%M}] {message_dict.get('user')}: {message_dict.get('text')}"
                text = message_format
        # there should only be Message objects passed into here, SOMEHOW we sometimes dont get one
        # or its 'text' attribute is "none"
        # TODO: this is a hack, needs investigation
        else:
            text = str(vl)
        return text

    def set_up_handlers(self):
        super(ChannelMessages, self).set_up_handlers()
        self.handlers.update({
            curses.KEY_LEFT: self.h_exit_left,
            curses.KEY_RIGHT: self.h_exit_right
        })


class BoxedChannelMessages(npyscreen.BoxTitle):
    _contained_widget = ChannelMessages

    def __init__(self, *args, **kwargs):
        self.name = 'Messages'
        super(BoxedChannelMessages, self).__init__(*args, **kwargs)
        self.current_channel = None
        self.typing_user_watchdog_thread = TypingUserWatchdogThread(widget=self)

    def buffer(self, *args, **kwargs) -> None:
        self.entry_widget.buffer(*args, **kwargs)
        self.display()

    def clear_buffer(self, *args, **kwargs) -> None:
        self.entry_widget.clear_buffer(*args, **kwargs)

    def set_channel(self, ch: Channel) -> None:
        """
        function to set title of box to channel name and display associated information
        """
        if ch is not None:
            self.current_channel = ch

        new_name = "Messages | {name}".format(name=ch.name)

        if ch.topic:
            topic = ch.topic.get('value')
        elif ch.purpose:
            topic = ch.purpose.get('value')
        else:
            topic = None
        if topic:
            new_name += " ({})".format(topic)

        if ch.is_private:
            new_name += " [PRIVATE]"

        self.name = new_name
        self.typing_user_watchdog_thread.start()

    def typing_user_event(self):
        try:
            typing_users = [u.get_name() for u in self.current_channel.typing_users.keys()]
        except:
            typing_users = []

        if len(typing_users) < 1:
            self.footer = None
        elif len(typing_users) == 1:
            self.footer = '{} is typing...'.format(typing_users[0])
        elif len(typing_users) < 4:
            self.footer = '{} and {} are typing...'.format(', '.join(typing_users[:-1]), typing_users[-1])
        elif len(typing_users) >= 4:
            self.footer = 'Multiple people are typing...'
        self.display()

    def destroy(self):
        self.typing_user_watchdog_thread.stop()
        super(BoxedChannelMessages, self).destroy()


class TypingUserWatchdogThread:
    def __init__(self, widget):
        self.widget = widget
        self.thread = threading.Thread(target=self.main_loop)
        self.thread.daemon = True
        self._continue = None
        self.running = False
        self.logger = Logger('')

    def main_loop(self):
        while self._continue:
            try:
                if self.widget.current_channel is not None:
                    if hasattr(self.widget.current_channel, 'typing_users'):
                        if self.widget.current_channel.typing_users is not None:
                            prev_len = len(self.widget.current_channel.typing_users.keys())
                            self.widget.current_channel.typing_users = {
                                u: t for u, t in self.widget.current_channel.typing_users.items() if time.time() < t + 5
                            }
                            if len(self.widget.current_channel.typing_users.keys()) < prev_len:
                                self.widget.typing_user_event()
                time.sleep(2)
            except Exception as e:
                self.logger.log(e.args)

    def start(self):
        if not self.running:
            self._continue = True
            self.thread.start()
            self.running = True

    def stop(self):
        if self.running:
            self._continue = False
            self.thread.join(timeout=3)
            self.running = False


import threading

from .channel import Channel
from .user import User
from slackclient import SlackClient


class SlackApiClient:
    PUBLIC = 'public_channel'
    PRIVATE = 'private_channel'
    IM = 'im'
    MPIM = 'mpim'

    def __init__(self, config):
        self.token: str = config.token
        self.slackclient: SlackClient = SlackClient(self.token)
        self.channels = {}
        self.users = {}
        self.refresh_user_list()
        self.refresh_channel_list()

    def refresh_channel_list(self) -> None:
        channels = self.get_my_channels(_type=self.PUBLIC)
        channels.sort(key=lambda c: c.name)
        self.channels = {str(c.id): c for c in channels}

        private_channels = self.get_my_channels(_type=self.PRIVATE)
        private_channels.sort(key=lambda c: c.name)
        self.channels.update({str(c.id): c for c in private_channels})

        im_channels = self.get_my_channels(_type=self.IM)
        im_channels.sort(key=lambda c: c.name)
        self.channels.update({str(c.id): c for c in im_channels})

    def get_my_channels(self, _type: str=None) -> list:
        def _channel_scraper_thread(item, res_list):
            res = Channel(self, **item)
            res_list.append(res)
            return res

        channels = {}
        if _type is None:
            types = (self.PUBLIC, self.PRIVATE, self.IM, self.MPIM)
        else:
            types = [_type]
        for t in types:
            response = self.slackclient.api_call('users.conversations',
                                                 types=t)
            if response.get('ok'):
                print('fetching channel info for type {}...'.format(t))
                threads = []
                res_list = []
                for item in response.get('channels'):
                    thread = threading.Thread(target=_channel_scraper_thread, args=[item, res_list])
                    threads.append(thread)
                    thread.start()
                for thread in threads:
                    thread.join()
                channels[t] = [r for r in res_list if r is not None]
        return channels.get(_type) if _type else channels

    def refresh_user_list(self) -> None:
        self.users = {str(u.id): u for u in self.get_users()}

    def get_active_channels(self) -> list:
        response = self.slackclient.api_call("channels.list", exclude_archived=1)
        if response.get('ok'):
            return [Channel(self, **item) for item in response.get('channels')]

    def get_active_channels_im_in(self) -> list:
        return list(self.channels.values())

    def get_users(self) -> list:
        response = self.slackclient.api_call('users.list')
        if response.get('ok'):
            return [User(r) for r in response.get('members')]

    def rtm_connect(self) -> str:
        response = self.slackclient.api_call('rtm.connect')
        if response.get('ok'):
            return response.get('url')



class User:
    ADMIN_USER_PREFIX = ''
    PREFER_REAL_NAME = False

    def __init__(self, kwargs):
        self.id = kwargs.get('id')
        self.name = kwargs.get('name')
        self.display_name = kwargs.get('profile').get('display_name_normalized')
        self.real_name = kwargs.get('profile').get('real_name_normalized')
        self.is_admin = kwargs.get('is_admin')

    def get_name(self) -> str:
        if self.PREFER_REAL_NAME:
            name = self.real_name or self.display_name
        else:
            name = self.display_name or self.real_name
        return (self.ADMIN_USER_PREFIX if self.is_admin else '') + name

    def __str__(self) -> str:
        return self.get_name()
